/*--------------------------------------------------------------------*/
/*--- Magic sequence management:                   		   ---*/
/*---       	init, query and registration of                    ---*/
/*--- 		authorized magic sequence users       		   ---*/
/*--- 								   ---*/
/*---                                  rise_magic_sequence_mgmnt.c ---*/
/*--------------------------------------------------------------------*/
/*
   This file is part of RISE, an instruction set randomizer for security,
   built on top of Valgrind, an extensible x86 protected-mode
   emulator for monitoring program execution on x86-Unixes.

   RISE:
   Copyright (C) 2002-2003 Gabriela Barrantes
   	gbarrant@ieee.org

   Valgrind:
   Copyright (C) 2000-2003 Julian Seward 
      jseward@acm.org

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.

   The GNU General Public License is contained in the file COPYING.
*/


#include "vg_include.h"
#include "rise_include.h"

// GBS: Data structures to authorize magic seq users
// What segments are considered "magic"?
//      - vgskin
//      - "libphthread" (valgrind's reimplementation of it)
//      - vg_main
// A "magic" user is authorized to use VG's magic sequence

// (defined in rise_include.h)
//typedef
//enum { VGSKIN, VGMAIN, VGTHREAD, numAreas}
//magicUsers;

/* The following structure is defined in
 * coregrind vg_include.h
 * but I am showing it here so the
 * init magic operations would be clear
 
typedef
struct _memarea {
        UInt start;
        UInt end;
} memarea;

*/

// (defined in rise_include.h)
//memarea authorized[numAreas];

void RISE_(init_magic_users)(void) {
        authorized[VGMAIN].start  = 0;
	authorized[VGMAIN].end    = 0;
        authorized[VGSKIN].start  = 0;
	authorized[VGSKIN].end    = 0;
        authorized[VGTHREAD].start  = 0;
	authorized[VGTHREAD].end    = 0;
}


//GBS: This function checks if an address is authorized
//to use magic sequences
Bool RISE_(authorized_to_use_magic_seq)(Addr eip) {


        if (( eip >= authorized[VGMAIN].start) && ( eip <= authorized[VGMAIN].end)) return True;
        if (( eip >= authorized[VGSKIN].start) && ( eip <= authorized[VGSKIN].end)) return True;
        if (( eip >= authorized[VGTHREAD].start) && ( eip <= authorized[VGTHREAD].end)) return True;

        return False;

}
 
//========================================================================================================
//
void RISE_(register_magic_area)(Addr start, UInt size) {
	int j;
     	j = 0;
	while ((j < numAreas) && (authorized[j].start != 0)) j++;
	if (j < numAreas){
		authorized[j].start = start;
		authorized[j].end = start + size - 1;
	} else
        VG_(core_panic)("Magic sequence users area full\n");
}

