/*--------------------------------------------------------------------*/
/*--- Scramble support operations: key creation,                   ---*/
/*--- retrieval and cleaning                                       ---*/
/*---                                              rise_scramble.c ---*/
/*--------------------------------------------------------------------*/
/*
   This file is part of RISE, an instruction set randomizer for security,
   built on top of Valgrind, an extensible x86 protected-mode
   emulator for monitoring program execution on x86-Unixes.

   RISE:
   Copyright (C) 2002-2003 Gabriela Barrantes
   	gbarrant@ieee.org

   Valgrind:
   Copyright (C) 2000-2003 Julian Seward 
      jseward@acm.org

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.

   The GNU General Public License is contained in the file COPYING.
*/

//GBS:Note to myself: if no further problems are detected with leaving
//urandom open, delete the commented parts

#include "vg_include.h"
#include "rise_include.h"

//GBS: Used in main to make sure that the pool of true randomness was
//"big enough" (in some sense) before using urandom

//(Defined in rise_include.h)
//#define seedSize 256


//GBS: RISE command line argument vars
Bool RISE_(noInstrumentation) = False;
Bool RISE_(noCacheProtection) = False;
Bool RISE_(printInrefs) = False;
UInt RISE_(mask_size) = 0;

UInt RISE_(num_masks) = 0;
UChar * * RISE_(allMasks);
int currentStatus;

//---------------***------------------------------------------------
void RISE_(ensureSeed)(void) {

// All commented-out to be able to run a real large number of crashes
/*
Int i,n_tot,fd;
UInt pseudoWait;

   fd = VG_(open)("/dev/random",VKI_O_RDONLY,0);
   n_tot = 0;
   
   i = VG_(ioctlRANDn)(fd,&n_tot);
   if (i == -1) VG_(core_panic)("Problems with /dev/random\n");
//   VG_(printf)("Seedsize (True Randomness) is %d bytes\n",n_tot);
   if (n_tot < seedSize) {
        VG_(printf)("===================================================================================\n");
        VG_(printf)("||  Your program will be paused until %d bytes of true randomness are generated ||\n",
                     seedSize);
        VG_(printf)("||  You can help to increase it by moving the mouse around                       ||\n");
        VG_(printf)("===================================================================================\n");
   }
   while ( n_tot < seedSize) {
	//GBS: This is probably overkill, but I wanted to give more time for events to happen
	for (pseudoWait=0; pseudoWait<100000;pseudoWait++);
        i = VG_(ioctlRANDn)(fd,&n_tot);
   }
   VG_(close)(fd);
*/
}


//---------------***------------------------------------------------
//(Defined in rise_include.h)
//GBS: File descriptor for /dev/urandom, to be used across the program
//Int  RISE_(urandomFD);

//---------------***------------------------------------------------
//GBS: /dev/urandom opener
void RISE_(open_urandom)(void) {
     RISE_(urandomFD) = VG_(open) ( "/dev/urandom", VKI_O_RDONLY, 0 );
     if (RISE_(urandomFD) == -1) {
         VG_(message)(Vg_UserMsg, "FATAL: can't open /dev/urandom");
         VG_(exit)(1);
     }
}

//---------------***------------------------------------------------
//GBS: /dev/urandom closer
void RISE_(close_urandom)(void) {
	VG_(close)(RISE_(urandomFD));
}


//---------------***------------------------------------------------
//GBS: Fills an area with random numbers from /dev/urandom
void RISE_(fill_with_random) (UChar * region, UInt size) 
{
  Int bytesRead;

  //VG_(printf)("Address received = %x\n",(Addr) region);

    /* Uncomment the following if having problems with an older 
     * Linux distro. I have had problems with RH 6.2*/
    /* RISE_(urandomFD) = VG_(open) ( "/dev/urandom", VKI_O_RDONLY, 0 );
      if (RISE_(urandomFD) == -1) {
      	VG_(message)(Vg_UserMsg, "FATAL: can't open /dev/urandom");
      	VG_(exit)(1);
      }
     */

  bytesRead = VG_(read) ( RISE_(urandomFD), &region[0], size);
  if (bytesRead != size) {
	  VG_(message)(Vg_UserMsg, "FATAL: I/O error on /dev/urandom" );
	  VG_(exit)(1);
	
  }

}


/*******************************************************************************/
/* The following structure is defined in
 *  coregrind vg_include.h
 *  but I am showing it here so the
 *  init emuls operations would be clear
 *
 *    typedef
 *    struct _memarea {
 *            UInt start;
 *            UInt end;
 *    } memarea;
 *
 */


/*******************************************************************************/

//GBS: Access procedure to the byte that contains the mask for a given 
//address -- original design idea by Dave Ackley
//Eventually there should be a design decision between 2 and 3 levels

//Defined in coregrind/vg_include.h
//#define MASK_BLOCK_SIZE 4096
//#define SIZE_LEVEL_0 256 
//#define SIZE_LEVEL_1 4096 
//#define SIZE_LEVEL_2 MASK_BLOCK_SIZE 
//#define MASK_BITS 12
//#define LEVEL_0_BITS 8 
//#define LEVEL_1_BITS 12
//#define LEVEL_2_BITS MASK_BITS 

//---------------***------------------------------------------------
//(Defined in rise_include.h)
//At level 0 we have pointers to pointers to chars
//UChar * * RISE_(level0)[SIZE_LEVEL_0]; 

//GBS: (this is insane... C actually allows for such constructs as
//Int ************z; 
//which means a pointer to a a pointer to a pointer ... to a pointer to an int.)

//---------------***------------------------------------------------
//GBS:Level 0 is used to direct references
//based on the highest part of a 32
//bits address

//GBS:Initializing in NULL all level0 entries
void RISE_(init_level0)(void) {
	UInt i;
	for (i=0;i<SIZE_LEVEL_0;i++) RISE_(level0)[i]= NULL;
}

//---------------***------------------------------------------------
//GBS:Initializing in NULL all entries of a level1 table
void init_level1_table(UChar * * L1table) {
	UInt i;
	for (i=0;i< SIZE_LEVEL_1;i++) 
		L1table[i] = NULL;
}

//---------------***------------------------------------------------
//GBS: Creates and initializes a level 1 table
UChar * * RISE_(createLevel1)(UInt tabadd) {

	// Check on tabadd integrity
	if (tabadd>=SIZE_LEVEL_0) //Not checking for < 0 because it's an unsigned int
		VG_(core_panic)("Attempting to access a level 0 entry out of range\n");

	// Check that we are  not creating a new keymap for
	// an existing one
	if (RISE_(level0)[tabadd] != NULL) 
		VG_(core_panic)("Attempting to create a level 1 table over an existing one\n");

	//Gets memory and connects it to the level 0 entry
	RISE_(level0)[tabadd] =  (UChar **) VG_(get_memory_from_mmap)(SIZE_LEVEL_1 * sizeof(UChar *),
		"Creating level 1 address table for mask allocation");

	//GBS: (I don't check for NULL because get_memory_from_mmap 
	//stops the emulation if it is unable to get the memory requested

	//Initialize the area with NULLs
	init_level1_table(RISE_(level0)[tabadd]); 

	return( RISE_(level0)[tabadd]);
}

//---------------***------------------------------------------------
//GBS:  If requested key size is less than total size,
//creates all "keys" at once and then uses them as necessary
//Notice: the big structure that gives access to the keys is still
//not linked to this structure in any way
void RISE_(createALLmasks)(void) {

	UInt allocationSize,i;

        // Level 1 table

        //Let's make the number of masks slightly bigger
	        if (RISE_(num_masks) == 1) RISE_(num_masks)++;
		
	//VG_(printf)("We are going to have %x masks\n",RISE_(num_masks));
	
	allocationSize = RISE_(num_masks) * sizeof(UChar *);
	RISE_(allMasks) = VG_(get_memory_from_mmap)(allocationSize,
			"Creating table for reduced mask allocation");
	for (i=0;i<RISE_(num_masks);i++) {
                UChar * x;
                x =     (UChar *) VG_(get_memory_from_mmap)(SIZE_LEVEL_2 * sizeof(UChar),
                                                "Mask area allocation for reduced masks");
                //VG_(printf)("pointer returned by mmap for entry %d = [%x]\n",
                //              i, (Addr) x);
                RISE_(allMasks)[i] =  x;
//		RISE_(allMasks)[i] =  (UChar *) VG_(get_memory_from_mmap)(SIZE_LEVEL_2 * sizeof(UChar),
//						"Mask area allocation for reduced masks"); 
		RISE_(fill_with_random) (RISE_(allMasks)[i], SIZE_LEVEL_2);
	        VG_(mprotect)(RISE_(allMasks)[i],SIZE_LEVEL_2,VKI_PROT_READ);

	}
}
//---------------***------------------------------------------------
//GBS: Chooses one of the pre-existing masks to be linked
UInt RISE_(whichOne)(void) {
	//Gets one random word
	Addr anAddr;
	//UInt entry;

	//GBS: This is an ugly way to create a small random number
	RISE_(fill_with_random) ((UChar *) &anAddr, sizeof(Addr));

	/* -- some uglier previous implementation
	      RISE_(fill_with_random) ((UChar *) &pointToChar, sizeof(UChar *));
	      VG_(printf)("We got %i\n", *pointToChar);
	      entry = rise_umoddi3(((UInt) (* pointToChar)),RISE_(num_masks));
	      VG_(printf)("Mask is %i\n", entry);
	*/

	//VG_(printf)("We got %x\n",anAddr);
        anAddr = rise_umoddi3(anAddr,RISE_(num_masks));
       //VG_(printf)("Mask is %d\n",anAddr);
       
       return anAddr;
}

//---------------***------------------------------------------------
//GBS: creates an unconnected clone
//     if zStart == 0 ==> starts with zeroes
//     if zStart == 1 ==> starts with random
//     aLimit gives the first byte of random (zStart == 0) or zeroes (zStart == 1)
//     
UChar * RISE_(createMaskClone)(UInt zStart, Addr aLimit) {
	UChar * clonePtr;	
	Int size1stHalf,size2ndHalf;

	if (aLimit > SIZE_LEVEL_2)
		VG_(core_panic)("Called to create a mask with illegal parameters\n");

	size1stHalf = aLimit;
	size2ndHalf = SIZE_LEVEL_2 - aLimit;
	
	//Create mask area and links it to the level 1 entry
	clonePtr =  (UChar *) VG_(get_memory_from_mmap)(SIZE_LEVEL_2 * sizeof(UChar),
						"Mask area allocation"); 
	//GBS: get_memory_from_mmap stops the emulation if it is unable
	//to get the memory requested

	//Fills the area with randomness
	if (zStart == 0) {
		//VG_(printf)("First half [%x - %x] with 0 (%d bytes)\n",0,aLimit-1,size1stHalf);
		//VG_(printf)("Second half [%x - %x] with RANDOM (%d bytes)\n",aLimit,SIZE_LEVEL_2-1,size2ndHalf);
		if (size1stHalf > 0)
			VG_(memset)(clonePtr, 0, size1stHalf);
		if (size2ndHalf > 0)
			RISE_(fill_with_random)(clonePtr+aLimit, size2ndHalf);
	} else if (zStart == 1) {
		//VG_(printf)("First half [%x - %x] with RANDOM (%d bytes)\n",0,aLimit-1,size1stHalf);
		//VG_(printf)("Second half [%x - %x] with 0 (%d bytes)\n",aLimit,SIZE_LEVEL_2-1,size2ndHalf);
		if (size1stHalf > 0)
			RISE_(fill_with_random)(clonePtr, size1stHalf);
		if (size2ndHalf > 0)
			VG_(memset)(clonePtr+aLimit, 0, size2ndHalf);
	} else {
		VG_(core_panic)("Called to create a mask with illegal parameters\n");
	}
		
	//DO NOT PROTECT HERE!!! -- wait until it's linked
       // VG_(mprotect)(clonePtr,SIZE_LEVEL_2,VKI_PROT_READ);

	return(clonePtr);
}

//---------------***------------------------------------------------
//GBS: creates a piece of the key
UChar * RISE_(createMask)(UChar * * level1, UInt maskEntry) {
		
	
	//Check that maskEntry make sense	
	if (maskEntry >= SIZE_LEVEL_1)
		VG_(core_panic)("Called to create a mask in an out of range level 1 entry\n");

	// Will not create a new keymap for an existing one
	if ( level1[maskEntry] != NULL) 
			VG_(core_panic)("Called to create a mask over an existing one\n");

        // Deal with the special case of a small number of masks
	if (RISE_(num_masks) != 0) {
		UInt which;
		which = RISE_(whichOne)();
		level1[maskEntry] =  RISE_(allMasks)[which];
		return(level1[maskEntry]);
	}

        // If you are here it's because there is a mask for every executable byt in
        // your code
	
	//Create mask area and links it to the level 1 entry
	level1[maskEntry] =  (UChar *) VG_(get_memory_from_mmap)(SIZE_LEVEL_2 * sizeof(UChar),
						"Mask area allocation"); 
	//GBS: get_memory_from_mmap stops the emulation if it is unable
	//to get the memory requested

	//Fills the area with randomness
	RISE_(fill_with_random) (level1[maskEntry], SIZE_LEVEL_2);

	//And protects it against writing
        VG_(mprotect)(level1[maskEntry],SIZE_LEVEL_2,VKI_PROT_READ);

	return(level1[maskEntry]);
}

//---------------***------------------------------------------------
//GBS: Returns the mask byte for a given address
__inline__ UChar getMask ( Addr eip) {

UInt 	high, med , low;
UChar 	* mask;
UChar  	* * level1;

	high 	= 	eip >> (LEVEL_1_BITS+LEVEL_2_BITS);
	med  	= 	((eip >> LEVEL_2_BITS) & (SIZE_LEVEL_1 - 1));
	low 	= 	eip & (SIZE_LEVEL_2 - 1);

	//VG_(printf)("getting mask for address %x\n",eip);
	//VG_(printf)("high byte %x\n",high);
	//VG_(printf)("med byte %x\n",med);
	//VG_(printf)("low byte %x\n",low);
	if ( (level1 = (UChar * *) RISE_(level0)[high]) == NULL) {
		level1 = (UChar * *) RISE_(createLevel1)(high);
	}

	//VG_(printf)("m1\n");
	if ( (mask =  (UChar *) level1[med]) == NULL) {
		mask = (UChar *) RISE_(createMask)(level1,med);
	}
	//VG_(printf)("m2\n");

	return (mask[low]);	
	// Debugging trick: this disables entirely the whole mask effort
	//return 0; 
}

//---------------***------------------------------------------------
//GBS: Very badly thought of function to regenerate keys for an
//unloaded library
void regenerateKey(Addr start, UInt size) {
/*
UInt ii,high, currentHigh, med, chun;
Addr *lev1;
UChar   * mask;
        //Because we are on an already sealed block, we
        //must unblock first and then write
        VG_(printf)("REMV:Regenerating key for %x-%x\n",start,start+size-1);
        ii = start;
        while (True) {
                high =  currentHigh = (ii >> (LEVEL_1_BITS+LEVEL_2_BITS));

                if ( ((Addr *) RISE_(level0)[high] ) != NULL) lev1 = (Addr *) RISE_(level0)[high];
                else VG_(core_panic)("G) REMV: Trying to re-key an inexistent key");

                while (high == currentHigh) {
                        med = ((ii >> LEVEL_2_BITS) & (SIZE_LEVEL_1 - 1));

                        if  (((UChar *) lev1[med]) != NULL) mask =  (UChar *) lev1[med];
                        else VG_(core_panic)("H) REMV: Trying to re-key an inexistent key");

                        VG_(mprotect)(mask,SIZE_LEVEL_2,VKI_PROT_READ|VKI_PROT_WRITE);
                        chun = start + size // (end)
                                        - ii; // (current positition)
                        if (chun > SIZE_LEVEL_2)
                                        // We have not reached the last block                               chun = SIZE_LEVEL_2 - (ii & (SIZE_LEVEL_2 - 1));
                        VG_(printf)("REMV: Filling with random %x-%x\n", (Addr) mask +
                                                        (Addr) (ii & (SIZE_LEVEL_2 - 1)),
                                                        (Addr) mask +
                                                        (Addr) (ii & (SIZE_LEVEL_2 - 1))
                                                        + chun - 1);
                        RISE_(fill_with_random)((UChar *) (
                                                        (Addr) mask +
                                                        (Addr) (ii & (SIZE_LEVEL_2 - 1))
                                                   ),
                                                chun  //Number of items
                                           );
                        VG_(mprotect)(mask,SIZE_LEVEL_2,VKI_PROT_READ);
                        ii= (ii & ~(SIZE_LEVEL_2 -1)) + SIZE_LEVEL_2; //Mask size increase
                        if (ii> (start + size -1))
                                break;
                        high    =  (ii >> (LEVEL_1_BITS+LEVEL_2_BITS));
                } // "while high == currentHigh"
                if (ii> (start + size -1))
                        break;
        } // "while true"
	*/
}

